@@ -1,7 +1,7 @@
 from .segmentor import Segmentor
 from .imspec import ImSpec
-from .vae import BaseVAE, VAE, rVAE
+from .dgm import BaseVAE, VAE, rVAE, jVAE, jrVAE
 from .loaders import load_model, load_ensemble
 
 __all__ = ["Segmentor", "ImSpec", "BaseVAE", "VAE", "rVAE",
-           "load_model", "load_ensemble"]
+           "jVAE", "jrVAE", "load_model", "load_ensemble"]
